"""Main entry point for the REPL."""

import sys

from colorama import init

from eryx.__init__ import CURRENT_VERSION
from eryx.frontend.parser import Parser
from eryx.runtime.environment import Environment
from eryx.runtime.runner import run_code

init(autoreset=True)


def start_repl(
    log_ast: bool = False, log_result: bool = False, log_tokens: bool = False
):
    """Start the REPL."""
    # Create the global scope
    environment = Environment()

    # Initialize the parser
    parser = Parser()

    # REPL
    print(f"\nEryx v{CURRENT_VERSION}")
    while True:
        try:
            # Accept input from the user
            source_code = input("> ")

            # Run the code
            result = run_code(
                source_code, log_ast, log_result, log_tokens, environment, parser
            )

            # Print the result
            if result:
                print(result)

        except KeyboardInterrupt:
            print()
            break

    sys.exit(0)


if __name__ == "__main__":
    start_repl()



"""Values and their types in the runtime environment."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Dict, List

if TYPE_CHECKING:
    from frontend.ast import Statement
    from runtime.environment import Environment


@dataclass()
class RuntimeValue:
    """Base class for all runtime values."""


@dataclass()
class NullValue(RuntimeValue):
    """Null value class."""

    value: None = None


@dataclass()
class NumberValue(RuntimeValue):
    """Number value class."""

    value: float


@dataclass()
class BooleanValue(RuntimeValue):
    """Boolean value class."""

    value: bool


@dataclass()
class ObjectValue(RuntimeValue):
    """Object value class."""

    properties: Dict[str, RuntimeValue]
    immutable: bool = False # TODO: Actually implement this


@dataclass()
class FunctionCall:
    """Function call class."""

    arguments: List[RuntimeValue]
    environment: "Environment"


@dataclass()
class NativeFunctionValue(RuntimeValue):
    """Native function value class."""

    call: Callable


@dataclass()
class FunctionValue(RuntimeValue):
    """Function value class."""

    name: str
    arguments: List[str]
    environment: "Environment"
    body: List["Statement"]


@dataclass()
class StringValue(RuntimeValue):
    """String value class."""

    value: str


@dataclass()
class ArrayValue(RuntimeValue):
    """Array value class."""

    elements: List[RuntimeValue]



"""Parser module for the frontend of the compiler."""

from eryx.frontend.ast import (
    ArrayLiteral,
    AssignmentExpression,
    BinaryExpression,
    CallExpression,
    Expression,
    FunctionDeclaration,
    Identifier,
    IfStatement,
    ImportStatement,
    MemberExpression,
    NumericLiteral,
    ObjectLiteral,
    Program,
    Property,
    ReturnStatement,
    Statement,
    StringLiteral,
    VariableDeclaration,
)
from eryx.frontend.lexer import Token, TokenType, tokenize
from eryx.utils.errors import syntax_error

# Precedence:
# 1. Member expression: obj.prop, obj[expr]
# 2. Call expression: func(), obj.method()
# 3. Exponentiation: **
# 4. Multiplication, division, modulo: *, /, %
# 5. Addition, subtraction: +, -
# 6. Bitwise operations: <<, >>, &, |, ^
# 7. Logical AND and OR: &&, ||
# 8. Comparison: ==, !=, >, >=, <, <=
# 9. Assignment: =


class Parser:
    """Parser class."""

    def __init__(self) -> None:
        self.source_code = ""
        self.tokens = []

    def not_eof(self) -> bool:
        """Check if the parser has not reached the end of the token stream."""
        return self.tokens[0].type != TokenType.EOF

    def at(self) -> Token:
        """Get the current token."""
        return self.tokens[0]

    def next(self) -> Token:
        """Get the current token and skip to next one (also called eat)."""
        return self.tokens.pop(0)

    def look_ahead(self, n: int) -> Token:
        """Look ahead n tokens."""
        return self.tokens[n]

    def assert_next(self, token_type: TokenType, error: str) -> Token:
        """Return the current token and assert that the next token is of a certain type."""
        token = self.next()
        if token.type != token_type:
            syntax_error(self.source_code, token.position, error)
        return token

    def parse_additive_expression(self) -> Expression:
        """Parse an additive expression."""
        left = self.parse_multiplicative_expression()

        while self.at().value in ("+", "-"):
            operator = self.next().value
            right = self.parse_multiplicative_expression()
            left = BinaryExpression(left, operator, right)

        return left

    def parse_call_member_expression(self) -> Expression:
        """Parse a call member expression."""
        member = self.parse_member_expression()

        if self.at().type == TokenType.OPEN_PAREN:
            return self.parse_call_expression(member)

        return member

    def parse_call_expression(self, caller: Expression) -> Expression:
        """Parse a call expression."""
        call_expression = CallExpression(self.parse_arguments(), caller)

        if self.at().type == TokenType.OPEN_PAREN:
            call_expression = self.parse_call_expression(call_expression)

        return call_expression

    def parse_arguments(self) -> list[Expression]:
        """Parse arguments."""
        self.assert_next(TokenType.OPEN_PAREN, "Expected an open parenthesis.")

        arguments = (
            []
            if self.at().type == TokenType.CLOSE_PAREN
            else self.parse_arguments_list()
        )

        self.assert_next(TokenType.CLOSE_PAREN, "Expected a closing parenthesis.")

        return arguments

    def parse_arguments_list(self) -> list[Expression]:
        """Parse an arguments list."""
        arguments = [self.parse_assignment_expression()]

        while self.at().type == TokenType.COMMA:
            self.next()  # Skip the comma
            arguments.append(self.parse_assignment_expression())

        return arguments

    def parse_member_expression(self) -> Expression:
        """Parse a member expression."""
        obj = self.parse_primary_expression()

        while self.at().type in (TokenType.OPEN_BRACKET, TokenType.DOT):
            operator = self.next()
            proprty = None
            computed = False

            if operator.type == TokenType.DOT:
                proprty = self.parse_primary_expression()  # Identifier

                if not isinstance(proprty, Identifier):
                    syntax_error(
                        self.source_code,
                        self.at().position,
                        "Expected an identifier as a property.",
                    )
            else:
                computed = True
                proprty = self.parse_expression()
                self.assert_next(TokenType.CLOSE_BRACKET, "Expected a closing bracket.")

            obj = MemberExpression(obj, proprty, computed)

        return obj

    def parse_multiplicative_expression(self) -> Expression:
        """Parse a multiplicative expression."""
        left = self.parse_exponentiation_expression()

        while self.at().value in ("/", "*", "%"):
            operator = self.next().value
            right = self.parse_exponentiation_expression()
            left = BinaryExpression(left, operator, right)

        return left

    def parse_exponentiation_expression(self) -> Expression:
        """Parse an exponentiation expression."""
        left = self.parse_call_member_expression()

        while self.at().value == "**":
            operator = self.next().value
            right = self.parse_call_member_expression()
            left = BinaryExpression(left, operator, right)

        return left

    def parse_array_expression(self) -> Expression:
        """Parse an array expression."""
        self.assert_next(
            TokenType.OPEN_BRACKET, "Expected an opening bracket '[' for array."
        )

        elements = []
        # Check for empty arrays
        if self.at().type != TokenType.CLOSE_BRACKET:
            elements.append(self.parse_expression())

            while self.at().type == TokenType.COMMA:
                self.next()  # Skip the comma
                elements.append(self.parse_expression())

        self.assert_next(
            TokenType.CLOSE_BRACKET, "Expected a closing bracket ']' for array."
        )
        return ArrayLiteral(elements)

    def parse_primary_expression(self) -> Expression:
        """Parse a primary expression."""
        token = self.at()

        match token.type:
            case TokenType.IDENTIFIER:
                return Identifier(self.next().value)
            case TokenType.NUMBER:
                return NumericLiteral(float(self.next().value))
            case TokenType.STRING:
                return StringLiteral(self.next().value)
            case TokenType.OPEN_BRACKET:
                return self.parse_array_expression()
            case TokenType.OPEN_BRACE:
                return self.parse_object_expression()
            case TokenType.OPEN_PAREN:
                self.next()  # Skip the open parenthesis
                expression = self.parse_expression()
                self.assert_next(
                    TokenType.CLOSE_PAREN,
                    "Unexpected token found inside parenthesised expression, " \
                        "expected closing parenthesis.",
                )  # Skip the close parenthesis
                return expression
            case _:
                syntax_error(self.source_code, token.position, "Unexpected token.")
                return Expression()  # This will never be reached

    def parse_assignment_expression(self) -> Expression:
        """Parse an assignment expression."""
        left = self.parse_comparison_expression()

        if (
            self.at().type == TokenType.EQUALS
            and self.look_ahead(1).type != TokenType.EQUALS
        ):
            self.next()  # Skip the equals sign

            value = self.parse_assignment_expression()
            return AssignmentExpression(left, value)

        return left

    def parse_expression(self) -> Expression:
        """Parse an expression."""
        return self.parse_assignment_expression()

    def parse_object_expression(self) -> Expression:
        """Parse an object expression."""

        self.next()  # Skip the open brace

        properties = []
        while self.not_eof() and self.at().type != TokenType.CLOSE_BRACE:
            key = self.assert_next(
                TokenType.IDENTIFIER, "Expected an identifier as a key."
            )

            if self.at().type == TokenType.COMMA:
                self.next()  # Skip the comma
                properties.append(Property(key.value))
                continue

            if self.at().type == TokenType.CLOSE_BRACE:
                properties.append(Property(key.value))
                continue

            self.assert_next(TokenType.COLON, "Expected a colon after the key.")

            value = self.parse_expression()
            properties.append(Property(key.value, value))

            if self.at().type != TokenType.CLOSE_BRACE:
                self.assert_next(
                    TokenType.COMMA,
                    "Expected a comma or closing brace after the value.",
                )

        self.assert_next(
            TokenType.CLOSE_BRACE, "Expected a closing brace after the object."
        )
        return ObjectLiteral(properties)

    def parse_comparison_expression(self) -> Expression:
        """Parse a comparison expression."""
        left = self.parse_logical_expression()

        while self.at().value in ("==", "!=", ">", ">=", "<", "<="):
            operator = self.next().value
            right = self.parse_logical_expression()
            left = BinaryExpression(left, operator, right)

        return left

    def parse_bitwise_expression(self) -> Expression:
        """Parse a bitwise expression."""
        left = self.parse_additive_expression()

        while self.at().value in ("&", "|", "^", "<<", ">>"):
            operator = self.next().value
            right = self.parse_additive_expression()
            left = BinaryExpression(left, operator, right)

        return left

    def parse_logical_expression(self) -> Expression:
        """Parse a logical expression."""
        left = self.parse_bitwise_expression()

        while self.at().value in ("&&", "||"):
            operator = self.next().value
            right = self.parse_bitwise_expression()
            left = BinaryExpression(left, operator, right)

        return left

    def parse_if_statement(self) -> Statement:
        """Parse an if statement."""
        self.next()  # Skip the if keyword
        self.assert_next(
            TokenType.OPEN_PAREN,
            "Expected an opening parenthesis for the if condition.",
        )

        condition = self.parse_expression()

        self.assert_next(
            TokenType.CLOSE_PAREN,
            "Expected a closing parenthesis after the if condition.",
        )

        self.assert_next(
            TokenType.OPEN_BRACE, "Expected an opening brace for the if body."
        )

        body = []
        while self.not_eof() and self.at().type != TokenType.CLOSE_BRACE:
            body.append(self.parse_statement())

        self.assert_next(
            TokenType.CLOSE_BRACE, "Expected a closing brace after the if body."
        )

        if self.at().type == TokenType.ELSE:
            self.next()
            self.assert_next(
                TokenType.OPEN_BRACE, "Expected an opening brace for the else body."
            )

            else_body = []
            while self.not_eof() and self.at().type != TokenType.CLOSE_BRACE:
                else_body.append(self.parse_statement())

            self.assert_next(
                TokenType.CLOSE_BRACE, "Expected a closing brace after the else body."
            )

            return IfStatement(condition, then=body, else_=else_body)

        return IfStatement(condition, then=body)

    def parse_return_statement(self) -> Statement:
        """Parse a return statement."""
        self.next()  # Skip the return keyword

        if self.at().type == TokenType.SEMICOLON:
            self.next()
            return ReturnStatement()

        value = self.parse_expression()

        self.assert_next(TokenType.SEMICOLON, "Expected a semicolon after the return.")

        return ReturnStatement(value)

    def parse_function_declaration(self) -> Statement:
        """Parse a function declaration."""
        self.next()  # Skip the func keyword
        name = self.assert_next(
            TokenType.IDENTIFIER,
            "Expected an function name after the function keyword.",
        ).value
        arguments = self.parse_arguments()

        parameters = []
        for argument in arguments:
            if isinstance(argument, Identifier):
                parameters.append(argument.symbol)
            else:
                syntax_error(
                    self.source_code,
                    self.at().position,
                    "Function arguments must be identifiers.",
                )

        self.assert_next(TokenType.OPEN_BRACE, "Expected an opening brace.")

        body = []
        while self.not_eof() and self.at().type != TokenType.CLOSE_BRACE:
            body.append(self.parse_statement())

        self.assert_next(
            TokenType.CLOSE_BRACE,
            "Expected a closing brace after the function body.",
        )

        return FunctionDeclaration(name, parameters, body)

    def parse_variable_declaration(self) -> Statement:
        """Parse a variable declaration."""
        is_constant = self.next().type == TokenType.CONST
        identifier = self.assert_next(
            TokenType.IDENTIFIER, "Expected an identifier after a declaration."
        ).value

        if self.at().type == TokenType.SEMICOLON:
            self.next()  # Skip the semicolon
            if is_constant:
                syntax_error(
                    self.source_code,
                    self.at().position,
                    "Constant declaration must have an initial value.",
                )

            return VariableDeclaration(is_constant, Identifier(identifier))

        self.assert_next(
            TokenType.EQUALS, "Expected an equals sign after the identifier."
        )

        declaration = VariableDeclaration(
            is_constant, Identifier(identifier), self.parse_expression()
        )

        self.assert_next(
            TokenType.SEMICOLON, "Expected a semicolon after the declaration."
        )

        return declaration

    def parse_import_statement(self) -> Statement:
        """Parse an import statement."""
        self.next()  # Skip the import keyword

        value = self.assert_next(
            TokenType.STRING, "Expected a string after the import keyword."
        )

        if self.at().type == TokenType.AS:
            self.next() # Skip the as keyword

            alias = self.assert_next(TokenType.STRING, "Expected a string for the import alias.")

            return ImportStatement(value.value, alias=alias.value)

        return ImportStatement(value.value)

    def parse_from_statement(self) -> Statement:
        """Parse a from statement."""
        self.next()  # Skip the from keyword

        from_value = self.assert_next(
            TokenType.STRING, "Expected a string after the from keyword."
        )

        if not self.at().type == TokenType.IMPORT:
            syntax_error(
                self.source_code,
                self.at().position,
                "Expected an import keyword after the from keyword.",
            )

        self.next()  # Skip the import keyword

        import_value = self.parse_array_expression()

        if not isinstance(import_value, ArrayLiteral):
            syntax_error(
                self.source_code,
                self.at().position,
                "Import properties must be an array.",
            )
            return ImportStatement(
                from_value.value, None
            )  # Type checker stuff, will never happen

        if not import_value.elements:
            syntax_error(
                self.source_code,
                self.at().position,
                "Import properties must not be empty.",
            )

        if not all(isinstance(item, StringLiteral) for item in import_value.elements):
            syntax_error(
                self.source_code,
                self.at().position,
                "Import properties must be strings.",
            )

        return ImportStatement(
            from_value.value,
            [
                item.value
                for item in import_value.elements
                if isinstance(item, StringLiteral)
            ],
        )

    def parse_statement(self) -> Statement:
        """Parse a statement."""
        match self.at().type:
            case TokenType.LET:
                return self.parse_variable_declaration()
            case TokenType.CONST:
                return self.parse_variable_declaration()
            case TokenType.FUNC:
                return self.parse_function_declaration()
            case TokenType.IF:
                return self.parse_if_statement()
            case TokenType.RETURN:
                return self.parse_return_statement()
            case TokenType.IMPORT:
                return self.parse_import_statement()
            case TokenType.FROM:
                return self.parse_from_statement()
            case _:
                return self.parse_expression()

    def produce_ast(self, source_code: str) -> Program:
        """Produce an abstract syntax tree (AST) from source code."""
        self.source_code = source_code
        self.tokens = tokenize(source_code)
        program = Program(body=[])

        # Parse all statements in the program until the EOF
        while self.not_eof():
            program.body.append(self.parse_statement())

        return program



"""Web IDE for Eryx."""

import io
import re
import time
import uuid
from contextlib import redirect_stdout

from colorama import Fore
from flask import Flask, abort, jsonify, render_template, request

from eryx.__init__ import CURRENT_VERSION
from eryx.frontend.lexer import tokenize
from eryx.frontend.parser import Parser
from eryx.runtime.environment import Environment, get_value
from eryx.runtime.interpreter import evaluate
from eryx.utils.pretty_print import pprint

app = Flask(__name__)
app.config["SEND_FILE_MAX_AGE_DEFAULT"] = 600
parser = Parser()
environments = {}


class Config:
    """Web IDE configuration class."""

    def __init__(self):
        self.disable_file_io = False


config = Config()

# https://stackoverflow.com/questions/19212665/python-converting-ansi-color-codes-to-html
COLOR_DICT = {
    "31": ["hotpink"],
    "32": ["limegreen"],
    "33": ["gold"],
    "34": ["cyan"],
    "35": ["hotpink"],
    "36": ["cyan"],
    "37": ["white"],
}

COLOR_REGEX = re.compile(r"\[(?P<arg_1>\d+)(;(?P<arg_2>\d+)(;(?P<arg_3>\d+))?)?m")

TEMPLATE = '<span style="color: {}">'


def ansi_to_html(text):
    """Format ANSI text with color codes to HTML while also escaping it."""

    text = escape_html(text)

    def single_sub(match):
        argsdict = match.groupdict()
        if argsdict["arg_3"] is None:
            if argsdict["arg_2"] is None:
                color = argsdict["arg_1"]
            else:
                color = argsdict["arg_1"]
        else:
            color = argsdict["arg_2"]

        if color == "39":
            return "</span>"

        color_name = COLOR_DICT[color][0]
        return TEMPLATE.format(color_name)

    return COLOR_REGEX.sub(single_sub, text).replace("\u001b", "")


# ========


class TokenList:
    """List of tokens to use with the pretty printer."""

    def __init__(self, tokens: list):
        self.tokens = tokens


def escape_html(text):
    """Escape a string"""
    escape_dict = {"&": "&amp;", "<": "&lt;", ">": "&gt;", '"': "&quot;", "'": "&apos;"}

    return "".join(escape_dict.get(c, c) for c in text)


def refresh_env_uuid(env_uuid):
    """Refresh the expiry time of an environment."""
    environments[env_uuid]["expiry"] = time.time() + 600


def get_unique_uuid(dictionary):
    """Get a unique UUID that does not exist in dictionary."""
    while True:
        uid = uuid.uuid4().hex
        if uid not in dictionary:
            return uid


@app.route("/")
def index():
    """Main page."""
    return render_template("index.html", version=CURRENT_VERSION)


@app.route("/<action>", methods=["POST"])
def handle_actions(action):
    """Handle all IDE actions."""
    if action not in ["tokenize", "ast", "run", "result"]:
        abort(404)

    request_json = request.get_json()
    source_code = request_json["source_code"]
    if action == "tokenize":
        output_buffer = io.StringIO()
        with redirect_stdout(output_buffer):
            try:
                tokens = tokenize(source_code)
            except (RuntimeError, SystemExit) as e:
                if isinstance(e, SystemExit):
                    return jsonify(
                        {"error": ansi_to_html(Fore.RED + output_buffer.getvalue())}
                    )
                return jsonify({"error": ansi_to_html(Fore.RED + str(e))})
            return jsonify(
                {"result": ansi_to_html(pprint(TokenList(tokens), print_output=False))}
            )

    output_buffer = io.StringIO()
    with redirect_stdout(output_buffer):
        try:
            ast_nodes = parser.produce_ast(source_code)
            if action == "ast":
                return jsonify(
                    {"result": ansi_to_html(pprint(ast_nodes, print_output=False))}
                )
        except (SyntaxError, RuntimeError, SystemExit) as e:
            if isinstance(e, SystemExit):
                return jsonify(
                    {"error": ansi_to_html(Fore.RED + output_buffer.getvalue())}
                )
            return jsonify({"error": ansi_to_html(Fore.RED + str(e))})
    try:
        env = None
        is_repl = False
        if "env_uuid" in request_json:
            env_uuid = request_json["env_uuid"]
            env = environments.get(env_uuid)
            is_repl = True
        env = env["env"] if env else Environment(disable_file_io=config.disable_file_io)
        output_buffer = io.StringIO()
        with redirect_stdout(output_buffer):
            result = evaluate(ast_nodes, env)
            if action == "result":
                return jsonify(
                    {"result": ansi_to_html(pprint(result, print_output=False))}
                )
        if action == "run":
            output = output_buffer.getvalue()
            if is_repl and not output:
                return jsonify({"result": ansi_to_html(get_value(result))})
            return jsonify({"result": ansi_to_html(output)})
    except RuntimeError as e:
        return jsonify({"error": ansi_to_html(Fore.RED + str(e))})

    return jsonify({})


@app.route("/repl", methods=["POST", "DELETE"])
def repl():
    """REPL route."""
    if request.method == "POST":
        environment = Environment(disable_file_io=config.disable_file_io)
        env_uuid = get_unique_uuid(environments)
        environments[env_uuid] = {
            "env": environment,
            "expiry": time.time() + 600,
        }  # 10 minutes
        return jsonify({"env_uuid": env_uuid})

    # If method is DELETE
    if "envId" not in request.args:
        return jsonify({"error": "No environment UUID provided"})
    env_uuid = request.args["envId"]
    if env_uuid in environments:
        del environments[env_uuid]
    return jsonify({})


@app.route("/static/<path:path>", methods=["GET"])
def static_route(path):
    """Static file route."""
    return app.send_static_file(path)


@app.route("/favicon.ico")
def favicon():
    """Serve the favicon."""
    return app.send_static_file("eryx.ico")


def start_ide(host: str = "0.0.0.0", port: int = 80, disable_file_io: bool = False):
    """Start the web IDE."""
    config.disable_file_io = disable_file_io
    app.run(host=host, port=port, debug=False, use_reloader=False)


@app.before_request
def before_req_handler():
    """Delete expired environments when a request is received."""
    for env_uuid, env_data in environments.items():
        if time.time() > env_data["expiry"]:
            del environments[env_uuid]


if __name__ == "__main__":
    start_ide()



"""Module for managing Eryx packages."""

import io
import json
import os
import shutil
import zipfile

import requests
import toml

packages_dir = os.path.dirname(os.path.realpath(__file__))


CFG_FILE = "packages.json"
INSTALLED_PACKAGES_LOC = "installed"
DEFAULT_SERVER = "https://eryx-packages.shymike.tech"


def get_config() -> dict:
    """Initialize the package manager."""
    config_path = os.path.join(packages_dir, CFG_FILE)
    if not os.path.exists(config_path):
        with open(config_path, "w", encoding="utf8") as file:
            file.write("{}")

    try:
        with open(config_path, "r", encoding="utf8") as file:
            config = json.load(file)
    except json.JSONDecodeError:
        config = {}

    return config


class Config:
    """Class for managing the configuration file."""

    def __init__(self):
        self.config = get_config()

        if not self.config.get("installed_packages"):
            self.config["installed_packages"] = {}

    def __getitem__(self, key):
        return self.config[key]

    def __setitem__(self, key, value):
        self.config[key] = value
        with open(os.path.join(packages_dir, CFG_FILE), "w", encoding="utf8") as file:
            json.dump(self.config, file)

    def save(self) -> None:
        """Save the configuration file."""
        with open(os.path.join(packages_dir, CFG_FILE), "w", encoding="utf8") as file:
            json.dump(self.config, file)


CONFIG = Config()


def unzip_safe_from_buffer(
    path: str, zip_buffer: bytes, max_files: int = 100, max_size: int = 10**7
):  # 10MB
    """Unzip a buffer safely from a buffer."""
    # Process the zip buffer
    with zipfile.ZipFile(io.BytesIO(zip_buffer), "r") as zip_ref:
        # Ensure that the number of files and the total uncompressed size are within limits
        file_count = len(zip_ref.infolist())
        if file_count > max_files:
            raise ValueError(
                f"Too many files in the zip archive: {file_count} > {max_files}"
            )

        total_size = sum(file.file_size for file in zip_ref.infolist())
        if total_size > max_size:
            raise ValueError(
                f"Total size of uncompressed files exceeds limit: {total_size} > {max_size}"
            )

        # Extract files if safe
        zip_ref.extractall(path)


def get_folder_size(path: str) -> int:
    """Get the size of a folder."""
    total_size = 0
    for dirpath, _, filenames in os.walk(path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += os.path.getsize(fp)
    return total_size


def install(package: str, server: str, upgrade: bool) -> None:
    """Install an Eryx package."""

    if not server.startswith("http://") and not server.startswith("https://"):
        server = "https://" + server

    if not server.endswith("/"):
        server += "/"

    print(f"Installing package '{package}'")

    if "@" not in package:
        package_name = package
        if CONFIG["installed_packages"] and package in CONFIG["installed_packages"]:
            if not upgrade:
                print(f"Package '{package}' already installed")
                return

            package_path = os.path.join(
                packages_dir, INSTALLED_PACKAGES_LOC, package_name
            )
            if upgrade and os.path.exists(package_path):
                shutil.rmtree(package_path)
                del CONFIG["installed_packages"][package_name]

            try:
                if not os.path.exists(
                    os.path.join(packages_dir, INSTALLED_PACKAGES_LOC)
                ):
                    os.mkdir(os.path.join(packages_dir, INSTALLED_PACKAGES_LOC))
            except FileNotFoundError:
                pass

        try:
            response = None
            response = requests.get(f"{server}download/{package}", timeout=5)
            response.raise_for_status()
            version = (
                response.url.split("/")[-1].split("?")[0].rsplit(".", 1)[0] or "error"
            )

            package_file = response.content
        except requests.RequestException as e:
            if response:
                if response.status_code == 404:
                    print(f"Package '{package}' not found")
                else:
                    print(f"Error downloading package '{package}': {e}")
            else:
                print("Error downloading package: ", e)
            return

    else:
        package_name, version = package.split("@")

        if CONFIG["installed_packages"] and package in CONFIG["installed_packages"]:
            if not upgrade:
                print(
                    f"Package '{package}' already installed "
                    f"(installed version: {CONFIG['installed_packages'][package_name]})"
                )
                return

            package_path = os.path.join(
                packages_dir, INSTALLED_PACKAGES_LOC, package_name
            )
            if upgrade and os.path.exists(package_path):
                shutil.rmtree(package_path)
                del CONFIG["installed_packages"][package_name]

            try:
                if not os.path.exists(
                    os.path.join(packages_dir, INSTALLED_PACKAGES_LOC)
                ):
                    os.mkdir(os.path.join(packages_dir, INSTALLED_PACKAGES_LOC))
            except FileNotFoundError:
                pass

        try:
            response = None
            response = requests.get(
                f"{server}download/{package_name}/{version}", timeout=5
            )
            response.raise_for_status()
            package_file = response.content
        except requests.RequestException as e:
            if response:
                if response.status_code == 404:
                    try:
                        versions_response = requests.get(
                            f"{server}api/versions/{package_name}", timeout=5
                        )
                        versions_response.raise_for_status()
                        versions = versions_response.json().get("versions", [])
                        print(
                            f"Version '{version}' not found for package '{package_name}'"
                        )
                        print("Available versions:", ", ".join(versions))
                    except requests.RequestException:
                        print(f"Package '{package_name}' not found")
                else:
                    print(f"Error downloading package '{package_name}@{version}': {e}")
            else:
                print("Error downloading package: ", e)
            return

    CONFIG["installed_packages"][package_name] = version
    CONFIG.save()

    if not os.path.exists(os.path.join(packages_dir, INSTALLED_PACKAGES_LOC)):
        os.mkdir(os.path.join(packages_dir, INSTALLED_PACKAGES_LOC))

    unzip_safe_from_buffer(
        os.path.join(packages_dir, INSTALLED_PACKAGES_LOC, package_name), package_file
    )

    print(f"Package '{package_name}@{version}' installed successfully")


def uninstall(package: str) -> None:
    """Uninstall an Eryx package."""

    if not CONFIG["installed_packages"]:
        print("No packages installed")
        return

    if "@" in package:
        package_name, version = package.split("@")
    else:
        package_name = package
        version = None

    print(f"Uninstalling package '{package}'")

    if package not in CONFIG["installed_packages"]:
        if not version:
            print(f"Package '{package}' not installed")
            return
        if (
            package_name in CONFIG["installed_packages"]
            and CONFIG["installed_packages"][package_name] != version
        ):
            print(f"Package '{package_name}@{version}' not installed")
            return

    del CONFIG["installed_packages"][package_name]
    CONFIG.save()

    if not os.path.exists(os.path.join(packages_dir, INSTALLED_PACKAGES_LOC)):
        os.mkdir(os.path.join(packages_dir, INSTALLED_PACKAGES_LOC))

    try:
        shutil.rmtree(os.path.join(packages_dir, INSTALLED_PACKAGES_LOC, package_name))
    except FileNotFoundError:
        pass

    print(f"Package '{package}' uninstalled successfully")


def list_packages() -> None:
    """List all installed packages."""
    if not CONFIG["installed_packages"]:
        print("No packages installed")
        return

    print("Installed packages:")
    for package in CONFIG["installed_packages"]:
        print(f"  {package}@{CONFIG["installed_packages"][package]}")


def upload_package(package_folder: str, server: str) -> None:
    """Upload a package to the server."""

    if not server.startswith("http://") and not server.startswith("https://"):
        server = "https://" + server

    if not server.endswith("/"):
        server += "/"

    key = get_key(server)

    print(f"Uploading package from '{package_folder}'")

    if not os.path.exists(package_folder):
        print(f"Directory '{package_folder}' does not exist")
        return

    if not os.path.isdir(package_folder):
        print(f"'{package_folder}' is not a directory")
        return

    # Perform file checks
    if not os.path.exists(os.path.join(package_folder, "package.toml")):
        print("Missing 'package.toml' file")
        return

    if not os.path.exists(os.path.join(package_folder, "main.eryx")):
        print("Missing 'main.eryx' directory (package entry point)")
        return

    if get_folder_size(package_folder) > 10**7:
        print("Unzipped package too large: >10MB")
        return

    with open(
        os.path.join(package_folder, "package.toml"), "r", encoding="utf8"
    ) as file:
        try:
            package_data = toml.load(file)
        except toml.TomlDecodeError:
            print("Error decoding 'package.toml'")
            return

    if not package_data.get("package"):
        print("Missing 'package' table in 'package.toml'")
        return

    package_data = package_data["package"]

    package_name = str(package_data.get("name"))
    if not package_name:
        print("Missing 'name' field in 'package.toml'")
        return

    if not package_name.isidentifier():
        print("Invalid package name, can only contain letters, numbers and underscores")
        return

    package_version = str(package_data.get("version"))
    if not package_version:
        print("Missing 'version' field in 'package.toml'")
        return

    if not package_data.get("description"):
        print("Missing 'description' field in 'package.toml'")
        return

    # Zip the package
    shutil.make_archive(os.path.join(packages_dir, "temp"), "zip", package_folder)

    with open(os.path.join(packages_dir, "temp.zip"), "rb") as package_file:
        files = {"package_file": package_file.read()}

        if len(files["package_file"]) > 10**7:
            print("Zipped package too large: >1MB")
            return

        # Upload the package
        try:
            response = None
            response = requests.post(
                f"{server}/api/upload",
                headers={"X-API-Key": str(key)},
                files=files,
                timeout=5,
            )
            response.raise_for_status()
            print(f"Package '{package_name}@{package_version}' uploaded successfully")
        except requests.RequestException:
            if response:
                if response.status_code in (400, 401, 403):
                    CONFIG["api_key"] = None
                    print("Invalid API key")
                else:
                    try:
                        print(response.json()["error"])
                    except json.JSONDecodeError:
                        print("Error uploading package: ", response.text)
            else:
                print("Error uploading package")

    os.remove(os.path.join(packages_dir, "temp.zip"))


def delete_package(package: str, server: str) -> None:
    """Delete a package from the server."""

    if not server.startswith("http://") and not server.startswith("https://"):
        server = "https://" + server

    if not server.endswith("/"):
        server += "/"

    key = get_key(server)

    while True:
        answer = input(
            f"Are you sure you want to delete '{package}'\n"
            "THIS ACTION IS PERMANENT AND CANNOT BE UNDONE\n(y/N): "
        )
        if answer.lower() in ["y", "yes"]:
            break
        if answer.lower() in ["n", "no", ""]:
            print("Deletion cancelled")
            return
        print("Invalid input")

    try:
        response = None
        response = requests.post(
            f"{server}/api/delete",
            headers={"X-API-Key": str(key), "Content-Type": "application/json"},
            data=json.dumps({"package": package}).encode("utf-8"),
            timeout=5,
        )
        response.raise_for_status()
        print(f"Package '{package}' deleted successfully")
    except requests.RequestException:
        if response:
            if response.status_code == 401:
                CONFIG["api_key"] = None
            try:
                print(response.json()["error"])
            except json.JSONDecodeError:
                print("Error uploading package: ", response.text)
        else:
            print("Error uploading package")
        return


def get_key(server: str) -> str | None:
    """Get and save the API key."""

    cfg = get_config()

    if cfg.get("api_key"):
        return cfg["api_key"]

    print(f"\nPlease visit the following url to get your API key:\n{server}dashboard")

    key = input("\nAPI Key: ")
    with open(os.path.join(packages_dir, CFG_FILE), "w", encoding="utf8") as file:
        cfg["api_key"] = key
        json.dump(cfg, file)

    return key



"""Stuff for handling errors."""

from colorama import Fore


def position_to_line_column(source_code: str, position: int) -> tuple[int, int]:
    """Convert a position to a line and column number."""
    # Get the substring up to the given position
    substring = source_code[:position]

    # Count the number of newline characters to determine the line
    line = substring.count("\n") + 1

    # Find the column by looking for the last newline
    last_newline_pos = substring.rfind("\n")
    column = position - last_newline_pos if last_newline_pos != -1 else position + 1

    return (line, column)


def get_line_strings(source_code: str, line: int) -> str:
    """Get the line string from the source code."""
    lines = source_code.split("\n")

    return lines[line - 1]


# TODO: fix this, its VERY broken
def syntax_error(
    source_code: str, pos: int | tuple[int, int], error_message: str
) -> None:
    """Handle a syntax error."""
    length = 1 if isinstance(pos, int) else (pos[1] - pos[0])
    current_line, current_col = position_to_line_column(
        source_code, pos if isinstance(pos, int) else pos[0]
    )
    line = get_line_strings(source_code, current_line)
    error_line = f"{Fore.CYAN}{str(current_line).rjust(3)}:{Fore.RESET} {line}"
    marker = "^" * (length + 1)
    line_number_size = len(str(current_line).rjust(3))
    positioned_marker = marker.rjust(current_col + line_number_size + 1 + len(marker))
    formatted_marker = f"{Fore.YELLOW}{positioned_marker}{Fore.RESET}"
    error_msg = f"{Fore.RED}SyntaxError{Fore.RESET}: {error_message}"
    # pylint pls dont crash this time
    raise SyntaxError(f"{error_line}\n{formatted_marker}\n{error_msg}")



"""
Module to pretty print a class instance for debugging. (the code is actual garbage but it works)
"""

import os

from colorama import Fore, init

if os.name == "nt":
    init(autoreset=True)


# https://stackoverflow.com/questions/395735/how-to-check-whether-a-variable-is-a-class-or-not
def isclass(cls):
    """Check if a variable is a class."""
    return str(type(cls)).startswith("<class") and hasattr(cls, "__weakref__")


def isenum(cls):
    """Check if a variable is an enum."""
    return str(type(cls)).startswith("<enum") and hasattr(cls, "__weakref__")


def isfunction(func):
    """Check if a variable is a function."""
    return str(type(func)) == "<class 'function'>"


COLOR_DICT = {
    "class": Fore.MAGENTA,
    "enum": Fore.CYAN,
    "key": Fore.BLUE,
    "str": Fore.GREEN,
    "int": Fore.YELLOW,
    "float": Fore.YELLOW,
    "bool": Fore.RED,
    "function": Fore.CYAN,
    "NoneType": Fore.RED,
}


def get_color(class_instance):
    """Helper function to get the color of a value type."""
    # Extra check for enums
    if isenum(class_instance):
        return COLOR_DICT["enum"]
    # Class colors
    return COLOR_DICT.get(type(class_instance).__name__, Fore.RESET)


def handle_array(
    val, use_color, use_newlines, indent, _tabs, is_tuple=False, is_set=False
):
    """Helper function to handle lists, tuples and sets."""
    string = "{" if is_set else "[" if not is_tuple else "("
    if use_newlines:
        string += f"{' '*(indent*(_tabs+1))}"

    for i, item in enumerate(val):
        if isclass(item):
            if i > 0:
                string += ","
                if use_newlines:
                    string += f"\n{' '*(indent*(_tabs+1))}"

            if use_newlines:
                string += f"\n{' '*(indent*(_tabs+1))}"

            string += str(
                pprint(
                    item,
                    print_output=False,
                    use_color=use_color,
                    use_newlines=use_newlines,
                    indent=indent,
                    _tabs=_tabs + 1,
                )
            )
        else:
            if i > 0:
                string += ","
                if not use_newlines:
                    string += " "

            if use_newlines:
                string += f"\n{' '*(indent*(_tabs+1))}"

            if isinstance(item, (list, tuple, set)):
                # If list, set or tuple, call handle_array
                string += handle_array(
                    item,
                    use_color,
                    use_newlines,
                    indent,
                    _tabs + 1,
                    is_tuple=isinstance(item, tuple),
                    is_set=isinstance(item, set),
                )
            elif isinstance(item, dict):
                # If dict, call handle_dict
                string += handle_dict(item, use_color, use_newlines, indent, _tabs + 1)
            elif isinstance(item, str):
                # If string, call handle_str
                string += handle_str(item, use_color)
            elif isfunction(item):
                # If function, add the function name
                string += (
                    get_color(item) + f"{val.__name__}" + Fore.RESET
                    if use_color
                    else "" + "()"
                )
            else:
                # Else add the value with color
                string += (
                    (get_color(item) if use_color else "")
                    + str(item)
                    + (Fore.RESET if use_color else "")
                )
    if use_newlines:
        string += f"\n{' '*(indent*(_tabs))}"
    string += "}" if is_set else "]" if not is_tuple else ")"
    return string


def handle_dict(val, use_color, use_newlines, indent, _tabs):
    """Helper function to handle dictionaries."""
    string = "{"
    if use_newlines:
        string += f"{' '*(indent*(_tabs+1))}"

    for i, (key, value) in enumerate(val.items()):
        if use_newlines:
            string += f"\n{' '*(indent*(_tabs+1))}"
        if use_color:
            string += COLOR_DICT["key"]
        string += f"{key}"
        if use_color:
            string += Fore.RESET
        string += ": "

        if use_color:
            string += get_color(value)

        if isclass(value):
            # If class, recursively call pprint
            string += str(
                pprint(
                    value,
                    print_output=False,
                    use_color=use_color,
                    use_newlines=use_newlines,
                    indent=indent,
                    _tabs=_tabs + 1,
                )
            )
        elif isinstance(value, (list, tuple, set)):
            # If list, set or tuple, call handle_array
            string += handle_array(
                value,
                use_color,
                use_newlines,
                indent,
                _tabs + 1,
                is_tuple=isinstance(value, tuple),
                is_set=isinstance(value, set),
            )
        elif isinstance(value, dict):
            # If dict, call handle_dict
            string += handle_dict(value, use_color, use_newlines, indent, _tabs + 1)
        elif isinstance(value, str):
            # If string, call handle_str
            string += handle_str(value, use_color)
        elif isfunction(val):
            # If function, add the function name
            string += (
                get_color(val) + f"{val.__name__}" + Fore.RESET
                if use_color
                else "" + "()"
            )
        else:
            # Else add the value with color
            string += (
                (get_color(value) if use_color else "")
                + str(value)
                + (Fore.RESET if use_color else "")
            )

        if use_color:
            string += Fore.RESET

        if i + 1 < len(val):
            string += ","
            if not use_newlines:
                string += " "

    string += f"\n{' '*(indent*(_tabs))}" + "}"
    return string


def handle_str(val, use_color):
    """Helper function to handle strings."""
    if use_color:
        return f'{get_color(val)}"{val}"{Fore.RESET}'
    return f'"{val}"'


def pprint(
    class_instance,
    print_output: bool = True,
    use_color: bool = True,
    use_newlines: bool = True,
    indent: int = 2,
    _tabs=0,
) -> str | None:
    """
    Pretty print a class instance.
    """

    if not isclass(class_instance):
        raise TypeError("Argument must be a class instance.")

    # Get the properties of the class (excluding dunder methods)
    properties = list(filter(lambda x: not x.startswith("__"), dir(class_instance)))
    string = ""
    # Add color for the class name
    if use_color:
        string += COLOR_DICT["class"]
    # Add the class name
    string += f"{type(class_instance).__name__}"

    # Reset color and add an opening parenthesis
    if use_color:
        string += Fore.RESET
    string += "("

    # Loop through the properties
    for n, class_property in enumerate(properties):
        # Get the value of the property
        val = getattr(class_instance, class_property)

        # Add newlines and indentation
        if use_newlines:
            string += f'\n{" "*(indent*(_tabs+1))}'

        # Add color
        if use_color:
            string += COLOR_DICT["key"]

        # Add the property name
        string += f"{class_property}"

        if use_color:
            string += Fore.RESET
        string += ": "

        # Add color for the value
        if use_color:
            string += get_color(val)

        if isclass(val):
            # If class, recursively call pprint
            string += str(
                pprint(
                    val,
                    print_output=False,
                    use_color=use_color,
                    use_newlines=use_newlines,
                    indent=indent,
                    _tabs=_tabs + 1,
                )
            )
        elif isinstance(val, (list, tuple, set)):
            # If list, set or tuple, call handle_array
            string += handle_array(
                val,
                use_color,
                use_newlines,
                indent,
                _tabs + 1,
                is_tuple=isinstance(val, tuple),
                is_set=isinstance(val, set),
            )
        elif isinstance(val, dict):
            # If dict, call handle_dict
            string += handle_dict(val, use_color, use_newlines, indent, _tabs + 1)
        elif isinstance(val, str):
            # If string, call handle_str
            string += handle_str(val, use_color)
        elif isfunction(val):
            # If function, add the function name
            string += (
                get_color(val) + f"{val.__name__}" + Fore.RESET
                if use_color
                else "" + "()"
            )
        else:
            # Else add the value with color
            string += (
                (get_color(val) if use_color else "")
                + str(val)
                + (Fore.RESET if use_color else "")
            )

        # Reset color
        if use_color:
            string += Fore.RESET

        # Add a comma if not the last property
        if n + 1 < len(properties):
            string += ","
            if not use_newlines:
                string += " "

    # Add newlines and indentation and close the class
    if use_newlines:
        string += f"\n{' '*(indent*_tabs)})"
    else:
        string += ")"

    # Print or return the string
    if not print_output:
        return string
    print(string)
    return None



"""Script to generate test files."""

import os
import sys
import tkinter as tk
from io import StringIO
from tkinter import messagebox

from eryx.__init__ import CURRENT_VERSION
from eryx.frontend.parser import Parser
from eryx.runtime.environment import Environment
from eryx.runtime.interpreter import evaluate
from eryx.utils.pretty_print import pprint

current_path = os.path.dirname(os.path.abspath(__file__))
os.makedirs(os.path.join(current_path, "test"), exist_ok=True)


def generate_ast(code):
    """Generate AST from code using the parser."""
    parser = Parser()
    return parser.produce_ast(code)


def capture_output(func, *args, **kwargs):
    """Capture output printed during the execution of the code."""
    old_stdout = sys.stdout
    sys.stdout = StringIO()

    try:
        func(*args, **kwargs)
        return sys.stdout.getvalue()
    finally:
        sys.stdout = old_stdout


def create_test_files(code, description, test_name):
    """Create the test files with actual values."""
    # Create the folder to store the test files
    test_folder = os.path.join(current_path, "test", test_name)
    if not os.path.exists(test_folder):
        os.makedirs(test_folder)

    # Create .eryx file (the code itself)
    with open(
        os.path.join(test_folder, f"{test_name}.eryx"), "w", encoding="utf8"
    ) as f:
        f.write(code)

    # Generate AST
    test_ast = generate_ast(code)
    with open(
        os.path.join(test_folder, f"{test_name}.eryx.ast"), "w", encoding="utf8"
    ) as f:
        try:
            f.write(str(pprint(test_ast, print_output=False, use_color=False)))
        except RuntimeError as e:
            print(f"Parser Error: {e}")
            return

    # Capture printed output
    try:
        output = capture_output(evaluate, test_ast, Environment())
    except RuntimeError as e:
        print(f"Runtime Error: {e}")
        return
    with open(
        os.path.join(test_folder, f"{test_name}.eryx.output"), "w", encoding="utf8"
    ) as f:
        f.write(output[:-1])

    # Create info.test file (Description and expected behavior)
    with open(os.path.join(test_folder, "test.info"), "w", encoding="utf8") as f:
        f.write(f"Version: {CURRENT_VERSION}\n")
        f.write(f"Name: {test_name}\n")
        f.write(f"Description: {description}")

    messagebox.showinfo("Success", f"Test files for {test_name} have been created.")


def on_create_test():
    """Handle the create test button click."""
    code = code_text.get("1.0", tk.END).strip()
    description = description_entry.get()
    test_name = name_entry.get()

    if not code or not test_name:
        messagebox.showerror("Error", "Please provide both code and test name.")
        return

    # Call the function to create the test files
    create_test_files(code, description, test_name)


# Set up the Tkinter GUI
root = tk.Tk()
root.title("Test File Generator")
root.resizable(False, False)

# Code input
code_label = tk.Label(root, text="Code:")
code_label.pack()
code_text = tk.Text(root, height=10, width=50)
code_text.pack()

# Test name
name_label = tk.Label(root, text="Test Name:")
name_label.pack()
name_entry = tk.Entry(root, width=50)
name_entry.pack()

# Description
description_label = tk.Label(root, text="Description:")
description_label.pack()
description_entry = tk.Entry(root, width=50)
description_entry.pack()

# Create button
create_button = tk.Button(root, text="Create Test Files", command=on_create_test)
create_button.pack(pady=(0, 10))

# Run the Tkinter event loop
root.mainloop()



"""Interpreter for the runtime."""

import json
import os

from eryx.frontend.ast import (
    ArrayLiteral,
    AssignmentExpression,
    BinaryExpression,
    CallExpression,
    FunctionDeclaration,
    Identifier,
    IfStatement,
    ImportStatement,
    MemberExpression,
    NumericLiteral,
    ObjectLiteral,
    Program,
    ReturnStatement,
    Statement,
    StringLiteral,
    VariableDeclaration,
)
from eryx.frontend.parser import Parser
from eryx.packages.packages import CFG_FILE, INSTALLED_PACKAGES_LOC, packages_dir
from eryx.runtime.environment import BUILTINS, Environment
from eryx.runtime.values import (
    ArrayValue,
    BooleanValue,
    FunctionValue,
    NativeFunctionValue,
    NullValue,
    NumberValue,
    ObjectValue,
    RuntimeValue,
    StringValue,
)
from eryx.utils.pretty_print import pprint


# Custom exception to manage returns
class ReturnException(Exception):
    """Dummy exception to manage return statements."""

    def __init__(self, value):
        self.value = value


# STATEMENTS
def eval_variable_declaration(
    declaration: VariableDeclaration, environment: Environment
) -> RuntimeValue:
    """Evaluate a variable declaration."""
    value = (
        evaluate(declaration.value, environment) if declaration.value else NullValue()
    )
    return environment.declare_variable(
        declaration.identifier.symbol, value, declaration.constant
    )


def eval_function_declaration(
    ast_node: FunctionDeclaration, environment: Environment
) -> RuntimeValue:
    """Evaluate a function declaration."""

    func = FunctionValue(
        name=ast_node.name,
        arguments=ast_node.arguments,
        environment=environment,
        body=ast_node.body,
    )

    return environment.declare_variable(ast_node.name, func, False)


def eval_program(program: Program, environment: Environment) -> RuntimeValue:
    """Evaluate a program."""
    last_evaluated = NullValue()

    try:
        for statement in program.body:
            last_evaluated = evaluate(statement, environment)
    except ReturnException as e:
        raise RuntimeError("Return statement found outside of a function.") from e

    return last_evaluated


def eval_if_statement(
    if_statement: IfStatement, environment: Environment
) -> RuntimeValue:
    """Evaluate an if statement."""
    condition = evaluate(if_statement.condition, environment)
    result = NullValue()

    if isinstance(condition, (BooleanValue, NumberValue, StringValue, NullValue)):
        if condition.value:
            for statement in if_statement.then:
                result = evaluate(statement, environment)
            return result

        if if_statement.else_:
            for statement in if_statement.else_:
                if statement:  # Type check stuff
                    result = evaluate(statement, environment)
            return result

    return NullValue()


def eval_import_statement(
    import_statement: ImportStatement, environment: Environment
) -> RuntimeValue:
    """Evaluate an import statement."""
    module_name = import_statement.module

    if module_name in BUILTINS:
        if module_name in ("file") and environment.disable_file_io:
            raise RuntimeError("File I/O is disabled, unable to import 'file'.")

        module = BUILTINS.get(module_name)
        if module:
            if import_statement.names:
                for name in import_statement.names:
                    if name in module.properties:
                        environment.declare_variable(
                            name, module.properties[name], True, overwrite=True
                        )
                    else:
                        raise RuntimeError(
                            f"Variable/function '{name}' not found in module '{module_name}'."
                        )
            else:
                name = import_statement.alias or module_name
                environment.declare_variable(name, module, True)
        else:
            raise RuntimeError(f"Error importing builtin '{module_name}'")
    else:
        if module_name.endswith(".eryx"):
            if not os.path.exists(module_name):
                raise RuntimeError(f"File '{module_name}.eryx' does not exist.")

            # Import the file
            file_path = module_name
            with open(file_path + ".eryx", "r", encoding="utf8") as file:
                source_code = file.read()
        else:
            try:
                cfg_file_path = os.path.join(packages_dir, CFG_FILE)
                with open(cfg_file_path, "r", encoding="utf8") as file:
                    cfg = json.load(file)
            except (FileNotFoundError, json.JSONDecodeError) as e:
                raise RuntimeError(f"Package '{module_name}' not found.") from e

            installed_packages = cfg.get("installed_packages", {})
            if not installed_packages or module_name not in installed_packages:
                raise RuntimeError(f"Package '{module_name}' not found.")

            package_path = os.path.join(
                packages_dir, INSTALLED_PACKAGES_LOC, module_name
            )
            if not os.path.exists(package_path):
                raise RuntimeError(f"Installed package '{module_name}' not found.")

            entrypoint = os.path.join(package_path, "main.eryx")
            if not os.path.exists(entrypoint):
                raise RuntimeError(
                    "Entrypoint 'main.eryx' not found in "
                    f"installed package '{module_name}'."
                )

            with open(entrypoint, "r", encoding="utf8") as file:
                source_code = file.read()

        # Run the code
        new_environment = Environment(parent_env=environment)
        parser = Parser()
        evaluate(parser.produce_ast(source_code), new_environment)

        if not import_statement.names:
            # Declare the imported object in the current environment
            import_obj = ObjectValue(new_environment.variables)
            name = import_statement.alias or module_name
            environment.declare_variable(name, import_obj, True)
        else:
            # Import only the specified variables/functions
            for name in import_statement.names:
                if name in new_environment.variables:
                    environment.declare_variable(
                        name, new_environment.variables[name], True, overwrite=True
                    )
                else:
                    raise RuntimeError(
                        f"Variable/function '{name}' not found in module '{module_name}'."
                    )

    return NullValue()


# EXPRESSIONS
def eval_binary_expression(
    binop: BinaryExpression, environment: Environment
) -> RuntimeValue:
    """Evaluate a binary expression."""
    left = evaluate(binop.left, environment)
    right = evaluate(binop.right, environment)

    if isinstance(left, NumberValue) and isinstance(right, NumberValue):
        if binop.operator in ["+", "-", "*", "/", "%"]:
            return eval_numeric_binary_expression(left, right, binop.operator)

        if binop.operator in ["==", "!=", "<", ">", "<=", ">="]:
            return BooleanValue(
                eval_numeric_comparison_expression(left, right, binop.operator)
            )

        if binop.operator == "**":
            return NumberValue(left.value**right.value)

        if binop.operator in ["^", "&", "|", "<<", ">>"]:
            return eval_numeric_bitwise_expression(left, right, binop.operator)

        if binop.operator in ["&&", "||"]:
            return eval_logical_expression(left, right, binop.operator)

        raise RuntimeError(f"Unknown binary operator {binop.operator}.")

    if binop.operator in ["&&", "||"]:
        if isinstance(left, (BooleanValue, StringValue)) and isinstance(
            right, (BooleanValue, StringValue)
        ):
            return eval_logical_expression(left, right, binop.operator)
        raise RuntimeError(
            "Expected boolean, string or number values for logical operators."
        )

    if binop.operator == "+":
        if isinstance(left, StringValue) and isinstance(right, StringValue):
            return StringValue(left.value + right.value)

        if isinstance(left, ArrayValue) and isinstance(right, ArrayValue):
            return ArrayValue(left.elements + right.elements)

        if isinstance(left, ObjectValue) and isinstance(right, ObjectValue):
            return ObjectValue({**left.properties, **right.properties})

        return NullValue()

    if binop.operator == "==":
        if isinstance(left, ArrayValue) and isinstance(right, ArrayValue):
            return BooleanValue(left.elements == right.elements)

        if isinstance(left, ObjectValue) and isinstance(right, ObjectValue):
            return BooleanValue(left.properties == right.properties)

        if isinstance(left, (FunctionValue, NativeFunctionValue)) and isinstance(
            right, (FunctionValue, NativeFunctionValue)
        ):
            return BooleanValue(left == right)

        if isinstance(
            left, (StringValue, NumberValue, BooleanValue, NullValue)
        ) and isinstance(right, (StringValue, NumberValue, BooleanValue, NullValue)):
            return BooleanValue(left.value == right.value)

        return BooleanValue(False)

    if binop.operator == "!=":
        if isinstance(left, ArrayValue) and isinstance(right, ArrayValue):
            return BooleanValue(left.elements != right.elements)

        if isinstance(left, ObjectValue) and isinstance(right, ObjectValue):
            return BooleanValue(left.properties != right.properties)

        if isinstance(left, (FunctionValue, NativeFunctionValue)) and isinstance(
            right, (FunctionValue, NativeFunctionValue)
        ):
            return BooleanValue(left != right)

        if isinstance(
            left, (StringValue, NumberValue, BooleanValue, NullValue)
        ) and isinstance(right, (StringValue, NumberValue, BooleanValue, NullValue)):
            return BooleanValue(left.value != right.value)

        return BooleanValue(True)

    return NullValue()


def eval_member_expression(
    member: MemberExpression, environment: Environment
) -> RuntimeValue:
    """Evaluate a member expression."""
    object_value = evaluate(member.object, environment)

    if isinstance(object_value, ObjectValue):
        if member.computed:
            property_value = evaluate(member.property, environment)
            if not isinstance(property_value, StringValue):
                raise RuntimeError("Expected a string as a property.")
            property_value = property_value.value
        else:
            if not isinstance(member.property, Identifier):
                raise RuntimeError("Expected an identifier as a property.")
            property_value = member.property.symbol

        return object_value.properties.get(property_value, NullValue())

    if isinstance(object_value, ArrayValue):
        if member.computed:
            property_value = evaluate(member.property, environment)
            if not isinstance(property_value, NumberValue):
                raise RuntimeError("Expected a number as an index.")

            return (
                object_value.elements[int(property_value.value)]
                if len(object_value.elements) > int(property_value.value)
                else NullValue()
            )

        raise RuntimeError("Expected a computed property for an array (number).")

    raise RuntimeError("Expected an object or array.")


def eval_numeric_binary_expression(
    left: NumberValue, right: NumberValue, operator: str
) -> NumberValue | NullValue:
    """Evaluate a binary expression with two parsed numeric operands (always numbers)."""
    match operator:
        case "+":
            return NumberValue(left.value + right.value)
        case "-":
            return NumberValue(left.value - right.value)
        case "*":
            return NumberValue(left.value * right.value)
        case "/":
            if right.value == 0:
                raise RuntimeError("Division by zero.")
            return NumberValue(left.value / right.value)
        case "%":
            return NumberValue(left.value % right.value)

    return NullValue()


def eval_numeric_comparison_expression(
    left: NumberValue, right: NumberValue, operator: str
) -> bool:
    """Evaluate a numeric comparison expression."""
    match operator:
        case "==":
            return left.value == right.value
        case "!=":
            return left.value != right.value
        case "<":
            return left.value < right.value
        case ">":
            return left.value > right.value
        case "<=":
            return left.value <= right.value
        case ">=":
            return left.value >= right.value

    return False


def eval_logical_expression(
    left: BooleanValue | StringValue | NumberValue | NullValue,
    right: BooleanValue | StringValue | NumberValue | NullValue,
    operator: str,
) -> BooleanValue | NullValue:
    """Evaluate a logical expression."""
    match operator:
        case "&&":
            return BooleanValue(bool(left.value) and bool(right.value))
        case "||":
            return BooleanValue(bool(left.value) or bool(right.value))

    return NullValue()


def eval_numeric_bitwise_expression(
    left: NumberValue, right: NumberValue, operator: str
) -> NumberValue | NullValue:
    """Evaluate a numeric binary expression."""
    match operator:
        case "^":
            return NumberValue(int(left.value) ^ int(right.value))
        case "&":
            return NumberValue(int(left.value) & int(right.value))
        case "|":
            return NumberValue(int(left.value) | int(right.value))
        case "<<":
            return NumberValue(int(left.value) << int(right.value))
        case ">>":
            return NumberValue(int(left.value) >> int(right.value))

    return NullValue()


def eval_object_expression(
    obj: ObjectLiteral, environment: Environment
) -> RuntimeValue:
    """Evaluate an object expression."""
    properties = {}

    for prop in obj.properties:
        if prop.value:
            properties[prop.key] = evaluate(prop.value, environment)
        else:
            # If the property does not have a value, look up the variable in the environment
            # So that { x } will be evaluated as { x: x }
            properties[prop.key] = environment.lookup_variable(prop.key)

    return ObjectValue(properties)


def eval_identifier(identifier: Identifier, environment: Environment) -> RuntimeValue:
    """Evaluate an identifier."""
    return environment.lookup_variable(identifier.symbol)


def eval_assignment_expression(
    node: AssignmentExpression, environment: Environment
) -> RuntimeValue:
    """Evaluate an assignment expression."""
    if not isinstance(node.assigne, Identifier):
        raise RuntimeError("Expected an identifier on the left side of an assignment.")

    return environment.assign_variable(
        node.assigne.symbol, evaluate(node.value, environment)
    )


def eval_call_expression(
    expression: CallExpression, environment: Environment
) -> RuntimeValue:
    """Evaluate a call expression."""
    arguments = [evaluate(arg, environment) for arg in expression.arguments]
    func = evaluate(expression.caller, environment)

    if isinstance(func, NativeFunctionValue):
        result = func.call(arguments, environment)
        return result

    if isinstance(func, FunctionValue):
        function_environment = Environment(func.environment)

        for i, function_argument in enumerate(func.arguments):
            if i >= len(arguments):  # Allow less arguments than expected
                function_environment.declare_variable(
                    function_argument, NullValue(), False
                )
            else:
                function_environment.declare_variable(
                    function_argument, arguments[i], False
                )

        # Evaluate the function body statement by statement
        try:
            for statement in func.body:
                evaluate(statement, function_environment)
        except ReturnException as ret:
            return ret.value

        return NullValue()

    raise RuntimeError("Cannot call a non-function value.")


# MAIN
def evaluate(ast_node: Statement | None, environment: Environment) -> RuntimeValue:
    """Evaluate an AST node."""
    if not ast_node:
        return NullValue()

    match ast_node:
        case NumericLiteral():
            return NumberValue(ast_node.value)
        case StringLiteral():
            return StringValue(ast_node.value)
        case ArrayLiteral():
            return ArrayValue(
                [evaluate(element, environment) for element in ast_node.elements]
            )
        case Identifier():
            return eval_identifier(ast_node, environment)
        case BinaryExpression():
            return eval_binary_expression(ast_node, environment)
        case AssignmentExpression():
            return eval_assignment_expression(ast_node, environment)
        case CallExpression():
            return eval_call_expression(ast_node, environment)
        case Program():
            return eval_program(ast_node, environment)
        case VariableDeclaration():
            return eval_variable_declaration(ast_node, environment)
        case FunctionDeclaration():
            return eval_function_declaration(ast_node, environment)
        case MemberExpression():
            return eval_member_expression(ast_node, environment)
        case ObjectLiteral():
            return eval_object_expression(ast_node, environment)
        case IfStatement():
            return eval_if_statement(ast_node, environment)
        case ReturnStatement():
            # Directly evaluate and raise ReturnException if it's a return statement
            value = evaluate(ast_node.value, environment)
            raise ReturnException(value)
        case ImportStatement():
            return eval_import_statement(ast_node, environment)
        case _:
            print("=== AST node ERROR ===")
            pprint(ast_node)
            raise RuntimeError("Unknown AST node.")



"""Lexer for the fronted."""

from enum import Enum, auto
from typing import Any, Union

from colorama import Fore, init

from eryx.utils.errors import syntax_error

init(autoreset=True)


class TokenType(Enum):
    """All token types in the language."""

    NUMBER = auto()
    IDENTIFIER = auto()
    STRING = auto()

    OPEN_PAREN = auto()
    CLOSE_PAREN = auto()
    OPEN_BRACE = auto()
    CLOSE_BRACE = auto()
    OPEN_BRACKET = auto()
    CLOSE_BRACKET = auto()

    DOUBLE_QUOTE = auto()

    BINARY_OPERATOR = auto()

    LET = auto()
    CONST = auto()
    FUNC = auto()
    IF = auto()
    ELSE = auto()
    RETURN = auto()

    IMPORT = auto()
    FROM = auto()
    AS = auto()

    EQUALS = auto()

    COMMA = auto()
    COLON = auto()
    SEMICOLON = auto()
    DOT = auto()

    EOF = auto()


class Token:
    """Token class."""

    def __init__(
        self, value: Any, token_type: TokenType, position: Union[int, tuple[int, int]]
    ):
        self.value = value
        self.type = token_type
        self.position = position

    def __repr__(self) -> str:
        return f'Token("{self.value}", {self.type.name}, {self.position})'


KEYWORDS = {
    "let": TokenType.LET,
    "const": TokenType.CONST,
    "func": TokenType.FUNC,
    "if": TokenType.IF,
    "else": TokenType.ELSE,
    "return": TokenType.RETURN,
    "import": TokenType.IMPORT,
    "from": TokenType.FROM,
    "as": TokenType.AS
}


def is_skipable(char: str) -> bool:
    """Check if a character is a skipable character."""
    return char in (
        " ",
        "\n",
        "\t",
        "\r",
    )  # Skip spaces, newlines, tabs, and carriage returns


def tokenize(source_code: str) -> list[Token]:
    """Tokenize the source code."""
    tokens = []
    source_size = len(source_code)
    src = list(source_code)
    comment = False

    while len(src) > 0:
        negative_num = False
        current_pos = source_size - len(src)

        if comment:
            if src[0] in ("\n", "\r", ";"):
                comment = False
            src.pop(0)
            continue

        single_char_tokens = {
            "(": TokenType.OPEN_PAREN,
            ")": TokenType.CLOSE_PAREN,
            "{": TokenType.OPEN_BRACE,
            "}": TokenType.CLOSE_BRACE,
            "[": TokenType.OPEN_BRACKET,
            "]": TokenType.CLOSE_BRACKET,
            "+": TokenType.BINARY_OPERATOR,
            "*": TokenType.BINARY_OPERATOR,
            "/": TokenType.BINARY_OPERATOR,
            "%": TokenType.BINARY_OPERATOR,
            "^": TokenType.BINARY_OPERATOR,
            ";": TokenType.SEMICOLON,
            ",": TokenType.COMMA,
            ":": TokenType.COLON,
            ".": TokenType.DOT,
        }

        # Check for single character tokens first
        if src[0] in single_char_tokens:
            token = src.pop(0)

            # Power operator
            if token == "*" and len(src) > 0 and src[0] == "*":
                src.pop(0)
                tokens.append(Token("**", TokenType.BINARY_OPERATOR, current_pos))
                continue

            tokens.append(Token(token, single_char_tokens[token], current_pos))
            continue

        # Check for comments
        if src[0] == "#":
            comment = True
            src.pop(0)
            continue

        if src[0] == ">" and len(src) > 1 and src[1] == ">":
            src.pop(0)
            src.pop(0)
            tokens.append(Token(">>", TokenType.BINARY_OPERATOR, current_pos))
            continue

        if src[0] == "<" and len(src) > 1 and src[1] == "<":
            src.pop(0)
            src.pop(0)
            tokens.append(Token("<<", TokenType.BINARY_OPERATOR, current_pos))
            continue

        if src[0] == "&":
            src.pop(0)
            if len(src) > 1 and src[1] == "&":
                src.pop(0)
                tokens.append(Token("&&", TokenType.BINARY_OPERATOR, current_pos))
            else:
                tokens.append(Token("&", TokenType.BINARY_OPERATOR, current_pos))
            continue

        if src[0] == "|":
            src.pop(0)
            if len(src) > 1 and src[1] == "|":
                src.pop(0)
                tokens.append(Token("||", TokenType.BINARY_OPERATOR, current_pos))
            else:
                tokens.append(Token("|", TokenType.BINARY_OPERATOR, current_pos))
            continue

        # If its not a single character token, check for negative numbers
        if src[0] == "-":
            if len(src) > 0 and src[1].isdigit():
                negative_num = True
            else:
                tokens.append(Token(src.pop(0), TokenType.BINARY_OPERATOR, current_pos))
                continue

        # If its a negative number, remove the negative sign
        if negative_num:
            src.pop(0)

        # Check for multi character tokens
        if src[0].isdigit():  # Number
            start_pos = current_pos
            end_pos = start_pos + (1 if negative_num else 0)
            number = src.pop(0)
            if negative_num:
                number = "-" + number
            dots = 0
            while len(src) > 0 and (src[0].isdigit() or src[0] == "."):
                if src[0] == ".":
                    dots += 1
                    if dots > 1:
                        break
                end_pos += 1
                number += src.pop(0)
            tokens.append(Token(number, TokenType.NUMBER, (start_pos, end_pos)))

        elif src[0].isalpha() or src[0] == "_":  # Identifier
            start_pos = current_pos
            end_pos = start_pos
            identifier = src.pop(0)
            while len(src) > 0 and (
                src[0].isalpha() or src[0].isdigit() or src[0] == "_"
            ):
                end_pos += 1
                identifier += src.pop(0)

            if identifier in KEYWORDS:
                tokens.append(
                    Token(identifier, KEYWORDS[identifier], (start_pos, end_pos))
                )
            else:
                tokens.append(
                    Token(identifier, TokenType.IDENTIFIER, (start_pos, end_pos))
                )

        elif is_skipable(src[0]):  # Skip spaces, newlines, tabs, and carriage returns
            src.pop(0)

        elif src[0] == '"':  # String
            start_pos = current_pos
            end_pos = start_pos
            src.pop(0)
            string = ""
            while len(src) > 0 and src[0] != '"':
                end_pos += 1
                string += src.pop(0)
            src.pop(0)
            tokens.append(Token(string, TokenType.STRING, (start_pos, end_pos + 1)))

        elif src[0] in ("=", "<", ">"):  # Binary operator
            if len(src) > 1:
                if src[0] == "=" and src[1] == "=":
                    tokens.append(
                        Token(
                            "==",
                            TokenType.BINARY_OPERATOR,
                            (current_pos, current_pos + 1),
                        )
                    )
                    src.pop(0)
                    src.pop(0)
                    continue

                if src[0] == "<" and src[1] == "=":
                    tokens.append(
                        Token(
                            "<=",
                            TokenType.BINARY_OPERATOR,
                            (current_pos, current_pos + 1),
                        )
                    )
                    src.pop(0)
                    src.pop(0)
                    continue

                if src[0] == ">" and src[1] == "=":
                    tokens.append(
                        Token(
                            ">=",
                            TokenType.BINARY_OPERATOR,
                            (current_pos, current_pos + 1),
                        )
                    )
                    src.pop(0)
                    src.pop(0)
                    continue

            if src[0] in ("<", ">"):
                tokens.append(Token(src.pop(0), TokenType.BINARY_OPERATOR, current_pos))
                continue

            if src[0] == "=":
                tokens.append(Token(src.pop(0), TokenType.EQUALS, current_pos))

        elif src[0] == "!" and len(src) > 1 and src[1] == "=":  # Binary operator
            tokens.append(
                Token("!=", TokenType.BINARY_OPERATOR, (current_pos, current_pos + 1))
            )
            src.pop(0)
            src.pop(0)

        else:
            # If this is reached, its an unknown character
            syntax_error(
                source_code,
                current_pos,
                f"Unknown character found in source '{Fore.MAGENTA}{src.pop(0)}{Fore.RESET}'",
            )

    tokens.append(Token("EOF", TokenType.EOF, source_size - len(src)))

    return tokens



"""Module for running Eryx code."""

from colorama import Fore

from eryx.frontend.lexer import tokenize
from eryx.frontend.parser import Parser
from eryx.runtime.environment import Environment, get_value
from eryx.runtime.interpreter import evaluate
from eryx.runtime.values import NullValue
from eryx.utils.pretty_print import pprint


class TokenList:
    """List of tokens to use with the pretty printer."""

    def __init__(self, tokens: list):
        self.tokens = tokens


def run_code(
    source_code: str,
    log_ast: bool = False,
    log_result: bool = False,
    log_tokens: bool = False,
    environment: Environment | None = None,
    parser: Parser | None = None,
) -> str | None:
    """Run an Eryx file."""

    result = None

    environment = environment or Environment()
    parser = parser or Parser()

    try:
        if log_tokens:
            try:
                tokenized = tokenize(source_code)
                print("Tokenized:")
                pprint(TokenList(tokenized))
            except RuntimeError as e:
                print(f"{Fore.RED}Tokenizer Error: {e}{Fore.RESET}")
                return

        try:
            ast = parser.produce_ast(source_code)
            if log_ast:
                print("AST:")
                pprint(ast)
        except RuntimeError as e:
            print(f"{Fore.RED}Parser Error: {e}{Fore.RESET}")
            return
    except SyntaxError as e:
        print(e)
        return

    try:
        result = evaluate(ast, environment)
        if log_result:
            print("\nResult:")
            pprint(result)
    except RuntimeError as e:
        print(f"{Fore.RED}Runtime Error: {e}{Fore.RESET}")

    if result is not None and not isinstance(result, NullValue):
        return get_value(result)
    return None



"""Script to run all tests using pytest."""

import os
import warnings

import pytest

from eryx.__init__ import CURRENT_VERSION
from eryx.frontend.parser import Parser
from eryx.runtime.environment import Environment
from eryx.runtime.interpreter import evaluate
from eryx.utils.pretty_print import pprint

current_path = os.path.dirname(os.path.abspath(__file__))
os.makedirs(os.path.join(current_path, "test"), exist_ok=True)
base_folder = os.path.join(current_path, "test")


def read_file(file_path: str) -> str:
    """Read a text file."""
    with open(file_path, "r", encoding="utf8") as file:
        return file.read()


def read_info(file_path: str) -> dict:
    """Read the info.test file to get the metadata."""
    with open(file_path, "r", encoding="utf8") as file:
        lines = file.readlines()
        info = {
            "version": lines[0].split(":")[1].strip(),
            "name": lines[1].split(":")[1].strip(),
            "description": lines[2].split(":")[1].strip(),
        }
    return info


@pytest.mark.parametrize(
    "test_folder",
    [
        f
        for f in os.listdir(os.path.join(current_path, "test"))
        if os.path.isdir(os.path.join(current_path, "test", f))
    ],
)
def test_eryx_code(test_folder: str, capfd: pytest.CaptureFixture):
    """Test Eryx code by parsing, producing the AST, evaluating it, and checking output."""

    environment = Environment()
    parser = Parser()

    # Get the paths to the test files
    eryx_code_path = os.path.join(base_folder, test_folder, f"{test_folder}.eryx")
    ast_expected_path = os.path.join(
        base_folder, test_folder, f"{test_folder}.eryx.ast"
    )
    output_expected_path = os.path.join(
        base_folder, test_folder, f"{test_folder}.eryx.output"
    )
    info_path = os.path.join(base_folder, test_folder, "test.info")

    # Read the info file
    info = read_info(info_path)

    # Version check
    if info["version"] != CURRENT_VERSION:
        warnings.warn(
            f"Test {test_folder} was made for version {info['version']} and may not work."
        )

    # Read the code
    test_code = read_file(eryx_code_path)

    # Step 1: Produce the AST
    test_ast = parser.produce_ast(test_code)

    expected_ast = read_file(ast_expected_path)
    assert (
        pprint(test_ast, use_color=False, print_output=False) == expected_ast
    ), f"AST for {test_folder} does not match expected result."

    # Step 2: Evaluate the AST
    evaluate(test_ast, environment)

    # Step 3: Check printed output
    captured = capfd.readouterr()
    expected_output = read_file(output_expected_path)
    assert (
        captured.out.strip() == expected_output
    ), f"Printed output for {test_folder} does not match expected output."


if __name__ == "__main__":
    pytest.main(["-v", __file__])



"""Module for lexer based syntax highlighting code."""

from colorama import Fore, init

from eryx.frontend.lexer import TokenType, tokenize

init(autoreset=True)

COLOR_TABLE = {
    TokenType.OPEN_PAREN: Fore.BLUE,
    TokenType.CLOSE_PAREN: Fore.BLUE,
    TokenType.OPEN_BRACE: Fore.BLUE,
    TokenType.CLOSE_BRACE: Fore.BLUE,
    TokenType.OPEN_BRACKET: Fore.BLUE,
    TokenType.CLOSE_BRACKET: Fore.BLUE,
    TokenType.BINARY_OPERATOR: Fore.YELLOW,
    TokenType.EQUALS: Fore.YELLOW,
    TokenType.SEMICOLON: Fore.YELLOW,
    TokenType.COMMA: Fore.YELLOW,
    TokenType.COLON: Fore.YELLOW,
    TokenType.DOT: Fore.YELLOW,
    TokenType.NUMBER: Fore.GREEN,
    TokenType.STRING: Fore.GREEN,
    TokenType.IDENTIFIER: Fore.RESET,
    TokenType.LET: Fore.RED,
    TokenType.CONST: Fore.RED,
    TokenType.FUNC: Fore.RED,
    TokenType.DOUBLE_QUOTE: Fore.GREEN,
    TokenType.EOF: Fore.RESET,
}

NATIVE_FUNCTIONS = [
    "print",
    "time",
    "input",
    "readfile",
    "len",
    "exit",
    "str",
    "int",
    "bool",
    "array",
    "type",
    "sum",
    "min",
    "max",
]


def highlight(source_code: str) -> str:
    """Highlight the source code."""
    tokens = tokenize(source_code)
    highlighted_code = list(source_code)
    offset = 0

    for token in tokens:
        if token.value in NATIVE_FUNCTIONS:
            color = Fore.CYAN
        else:
            color = COLOR_TABLE.get(token.type, Fore.RESET)

        if isinstance(token.position, int):
            highlighted_code.insert(token.position + offset, color)
            highlighted_code.insert(token.position + offset + 2, Fore.RESET)
        else:
            highlighted_code.insert(token.position[0] + offset, color)
            highlighted_code.insert(token.position[1] + offset + 2, Fore.RESET)
        offset += 2

    return "".join(highlighted_code)



"""Eryx entry point and Command Line Interface (CLI) module."""

import argparse
import os

import pytest
from colorama import init

from eryx.__init__ import CURRENT_VERSION
from eryx.packages.packages import (
    delete_package,
    install,
    list_packages,
    uninstall,
    upload_package,
)
from eryx.runtime.repl import start_repl
from eryx.runtime.runner import run_code
from eryx.server.ide import start_ide
from eryx.packages.packages import DEFAULT_SERVER

init(autoreset=True)
current_path = os.path.dirname(os.path.abspath(__file__))


def main():
    """CLI entry point."""
    arg_parser = argparse.ArgumentParser(
        description="Eryx Command Line Interface",
    )
    arg_parser.add_argument(
        "--version",
        action="version",
        version=f"Eryx, version {CURRENT_VERSION}",
        help="Show the version number and exit.",
    )

    # Set the program name if executed as a module
    if arg_parser.prog == "__main__.py":
        arg_parser.prog = "python -m eryx"

    # Create subparsers for multiple commands
    subparsers = arg_parser.add_subparsers(dest="command", help="Available commands")

    # 'repl' command
    repl_parser = subparsers.add_parser("repl", help="Start the REPL")
    repl_parser.add_argument(
        "--ast", action="store_true", help="Print the abstract syntax tree (AST)."
    )
    repl_parser.add_argument(
        "--result",
        action="store_true",
        help="Print the result of the evaluation.",
    )
    repl_parser.add_argument(
        "--tokenize", action="store_true", help="Print the tokenized source code."
    )

    # 'run' command
    run_parser = subparsers.add_parser("run", help="Run an Eryx file")
    run_parser.add_argument("filepath", type=str, help="File path to run.")
    run_parser.add_argument(
        "--ast", action="store_true", help="Print the abstract syntax tree (AST)."
    )
    run_parser.add_argument(
        "--result",
        action="store_true",
        help="Print the result of the evaluation.",
    )
    run_parser.add_argument(
        "--tokenize", action="store_true", help="Print the tokenized source code."
    )

    # 'server' command
    server_parser = subparsers.add_parser("server", help="Start the web IDE")
    server_parser.add_argument(
        "--port", type=int, help="Port number for the web IDE.", default=80
    )
    server_parser.add_argument(
        "--host", type=str, help="Host for the web IDE.", default="0.0.0.0"
    )
    server_parser.add_argument(
        "--no-file-io", action="store_true", help="Disable file I/O", default=False
    )

    # 'test' command
    subparsers.add_parser("test", help="Run the test suite")

    package_parser = subparsers.add_parser("package", help="Manage Eryx packages")
    package_subparsers = package_parser.add_subparsers(
        dest="package_command", help="Available package commands"
    )
    install_parser = package_subparsers.add_parser("install", help="Install a package")
    install_parser.add_argument("package", type=str, help="Package to install")
    install_parser.add_argument(
        "--upgrade", action="store_true", help="Upgrade package", default=False
    )
    install_parser.add_argument(
        "--server",
        type=str,
        help="Server to use",
        default=DEFAULT_SERVER,
    )

    uninstall_parser = package_subparsers.add_parser("uninstall", help="Uninstall a package")
    uninstall_parser.add_argument("package", type=str, help="Package to uninstall")

    package_subparsers.add_parser("list", help="List all installed packages")

    upload_parser = package_subparsers.add_parser("upload", help="Upload a package")
    upload_parser.add_argument("package_folder", type=str, help="Package folder to upload")
    upload_parser.add_argument(
        "--server",
        type=str,
        help="Server to use",
        default=DEFAULT_SERVER,
    )

    delete_parser = package_subparsers.add_parser("delete", help="Delete a package")
    delete_parser.add_argument("package", type=str, help="Package to delete")
    delete_parser.add_argument(
        "--server",
        type=str,
        help="Server to use",
        default=DEFAULT_SERVER,
    )

    # Parse the arguments
    args = arg_parser.parse_args()

    # Handling each command
    if args.command == "repl":
        start_repl(log_ast=args.ast, log_result=args.result, log_tokens=args.tokenize)
    elif args.command == "run":
        try:
            with open(args.filepath, "r", encoding="utf8") as file:
                source_code = file.read()
            run_code(
                source_code,
                log_ast=args.ast,
                log_result=args.result,
                log_tokens=args.tokenize,
            )
        except FileNotFoundError as e:
            print(
                f"eryx: can't open file '{args.filepath}': [Errno {e.args[0]}] {e.args[1]}"
            )
    elif args.command == "server":
        start_ide(
            args.host,
            port=args.port,
            disable_file_io=args.no_file_io,
        )
    elif args.command == "test":
        pytest.main(["-v", os.path.join(current_path, "tests", "run_tests.py")])
    elif args.command == "package":
        try:
            if args.package_command == "install":
                install(args.package, args.server, args.upgrade)
            elif args.package_command == "uninstall":
                uninstall(args.package)
            elif args.package_command == "list":
                list_packages()
            elif args.package_command == "upload":
                upload_package(args.package_folder, args.server)
            elif args.package_command == "delete":
                delete_package(args.package, args.server)
            else:
                package_parser.print_help()
        except KeyboardInterrupt:
            print("\nOperation cancelled.")
    elif args.command is None:
        arg_parser.print_help()
    else:
        print(f"Unknown command: {args.command}")


if __name__ == "__main__":
    main()



"""Abstract syntax tree (AST) for the frontend."""

from dataclasses import dataclass, field
from typing import List, Union


@dataclass()
class Statement:
    """Base class for all statements in the AST."""


@dataclass()
class Program(Statement):
    """Program class."""

    body: List[Statement]


@dataclass()
class Expression(Statement):
    """Expression base class."""


@dataclass()
class AssignmentExpression(Expression):
    """Assignment expression class."""

    assigne: Expression
    value: Expression


@dataclass()
class BinaryExpression(Expression):
    """Binary expression class."""

    left: Expression
    operator: str
    right: Expression


@dataclass()
class Identifier(Expression):
    """Identifier class."""

    symbol: str


@dataclass()
class VariableDeclaration(Statement):
    """Variable declaration class."""

    constant: bool
    identifier: Identifier
    value: Union[Expression, None] = None


@dataclass()
class NumericLiteral(Expression):
    """Numeric literal class."""

    value: float


@dataclass()
class StringLiteral(Expression):
    """String literal class."""

    value: str


@dataclass()
class Property(Expression):
    """Property class."""

    key: str
    value: Union[Expression, None] = None


@dataclass()
class ObjectLiteral(Expression):
    """Object literal class."""

    properties: List[Property]


@dataclass()
class ArrayLiteral(Expression):
    """Represents an array literal."""

    elements: List[Expression]


@dataclass()
class CallExpression(Expression):
    """Binary expression class."""

    arguments: List[Expression]
    caller: Expression


@dataclass()
class MemberExpression(Expression):
    """Binary expression class."""

    object: Expression
    property: Expression
    computed: bool


@dataclass()
class FunctionDeclaration(Statement):
    """Function declaration class."""

    name: str
    arguments: List[str]
    body: list[Statement]


@dataclass()
class IfStatement(Statement):
    """If statement class."""

    condition: Expression
    then: list[Statement]
    else_: list[Union[Statement, None]] = field(default_factory=list)


@dataclass()
class ReturnStatement(Statement):
    """Return statement class."""

    value: Union[Expression, None] = None

@dataclass()
class ImportStatement(Statement):
    """Import statement class."""

    module: str
    names: List[str] | None = None
    alias: str | None = None



"""Environment class for storing variables (also called scope)."""

import math
import random
import sys
import time
from urllib.request import Request, urlopen

from eryx.runtime.values import (
    ArrayValue,
    BooleanValue,
    FunctionValue,
    NativeFunctionValue,
    NullValue,
    NumberValue,
    ObjectValue,
    RuntimeValue,
    StringValue,
)

BUILTINS = {}


# pylint: disable=invalid-name
class Environment:
    """Environment class."""

    def __init__(
        self, parent_env: "Environment | None" = None, disable_file_io: bool = False
    ):
        self.is_global = parent_env is None
        self.parent = parent_env
        self.constants = []
        self.variables = {}
        self.disable_file_io = (
            disable_file_io if not parent_env else parent_env.disable_file_io
        )

        if self.is_global:
            self.setup_scope()

    def declare_variable(
        self,
        variable_name: str,
        value: RuntimeValue,
        constant: bool = False,
        overwrite: bool = False,
    ) -> RuntimeValue:
        """Declare a variable in the current scope."""
        # Raise an exception if the variable is already declared
        if variable_name in self.variables:
            if not overwrite:
                raise RuntimeError(f'Variable "{variable_name}" already declared')

        self.variables[variable_name] = value

        if constant:
            self.constants.append(variable_name)

        return value

    def assign_variable(self, variable_name: str, value: RuntimeValue) -> RuntimeValue:
        """Assign a value to a variable in the current scope."""
        environment = self.resolve(variable_name)

        if variable_name in environment.constants:
            raise RuntimeError(f'Cannot assign to constant variable "{variable_name}"')

        environment.variables[variable_name] = value
        return value

    def lookup_variable(self, variable_name: str) -> RuntimeValue:
        """Lookup a variable in the current scope."""
        environment = self.resolve(variable_name)
        return environment.variables[variable_name]

    def resolve(self, variable_name: str) -> "Environment":
        """Resolve a variable name to an environment."""
        # Return self if variable_name exists in the current scope
        if variable_name in self.variables:
            return self
        # If it does not exist, check the parent scope
        if self.parent:
            return self.parent.resolve(variable_name)
        # If it does not exist in the parent scope, raise an exception
        raise RuntimeError(f'Variable "{variable_name}" not found in scope')

    def setup_scope(self) -> None:
        """Setup the global scope."""
        # Declare global variables
        self.declare_variable("true", BooleanValue(True), True)
        self.declare_variable("false", BooleanValue(False), True)
        self.declare_variable("null", NullValue(), True)

        # Declare native methods
        self.declare_variable("print", NativeFunctionValue(_print), True)
        self.declare_variable("input", NativeFunctionValue(_input), True)
        self.declare_variable("len", NativeFunctionValue(_len), True)
        self.declare_variable("exit", NativeFunctionValue(_exit), True)
        self.declare_variable("str", NativeFunctionValue(_str), True)
        self.declare_variable("int", NativeFunctionValue(_int), True)
        self.declare_variable("bool", NativeFunctionValue(_bool), True)
        self.declare_variable("array", NativeFunctionValue(_array), True)
        self.declare_variable("type", NativeFunctionValue(_type), True)


def get_value(value: RuntimeValue, inside_array: bool = False) -> str:
    """Get the value of a RuntimeValue."""
    result = ""

    if isinstance(value, NullValue):
        result = "null"

    elif isinstance(value, BooleanValue):
        result = str(value.value).lower()

    elif isinstance(value, NumberValue):
        result = (
            str(value.value)
            if int(value.value) != value.value
            else str(int(value.value))
        )

    elif isinstance(value, StringValue):
        if inside_array:
            result = '"' + value.value + '"'
        else:
            result = value.value

    elif isinstance(value, NativeFunctionValue):
        result = f"<native function {value.call.__name__[1:]}>"

    elif isinstance(value, FunctionValue):
        result = f"<function {value.name}>"

    elif isinstance(value, ArrayValue):
        result += "[ "
        for val in value.elements:
            result += f"{get_value(val, inside_array=True)}, "
        result = result[:-2] + " ]"

    elif isinstance(value, ObjectValue):
        result += "{ "
        for key, val in value.properties.items():
            result += f"{key}: {get_value(val, inside_array=True)}, "
        result = result[:-2] + " }"

    else:
        result = str(value)

    return result


# Native functions
def _print(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    values = []
    for arg in args:
        values.append(get_value(arg))
    print(*values)
    return NullValue()


def _sqrt(args: list[RuntimeValue], _: Environment):
    if not args:
        raise RuntimeError("Missing number value")
    if not isinstance(args[0], NumberValue):
        raise RuntimeError("Input type must be a number")
    return NumberValue(args[0].value ** 0.5)


def _random(_: list[RuntimeValue], __: Environment):
    return NumberValue(random.random())


def _time(_: list[RuntimeValue], __: Environment) -> RuntimeValue:
    return NumberValue(time.time())


def _input(args: list[RuntimeValue], __: Environment) -> RuntimeValue:
    if args and isinstance(args[0], StringValue):
        result = input(args[0].value)
    else:
        result = input()
    return StringValue(result)


def _getRequest(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing URL argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("URL must be a string")
    try:
        request = Request(args[0].value, method="GET")
        with urlopen(request) as response:
            return ObjectValue(
                {
                    "data": StringValue(response.read().decode("utf-8")),
                    "status": NumberValue(response.status),
                }
            )
    except Exception as e:
        raise RuntimeError(f"Error fetching URL '{args[0].value}'") from e


def _postRequest(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing URL or data argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("URL must be a string")
    if not isinstance(args[1], StringValue):
        raise RuntimeError("Data must be a string")
    try:
        request = Request(args[0].value, method="POST")
        request.add_header("Content-Type", "application/json")
        with urlopen(request, data=args[1].value.encode("utf-8")) as response:
            return ObjectValue(
                {
                    "data": StringValue(response.read().decode("utf-8")),
                    "status": NumberValue(response.status),
                }
            )
    except Exception as e:
        raise RuntimeError(f"Error fetching URL '{args[0].value}'") from e


def _readFile(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing filename argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Filename must be a string")
    try:
        with open(args[0].value, "r", encoding="utf8") as file:
            return StringValue(file.read())
    except FileNotFoundError as e:
        raise RuntimeError(f"File '{args[0].value}' not found") from e


def _writeFile(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing filename or content argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Filename must be a string")
    if not isinstance(args[1], StringValue):
        raise RuntimeError("Content must be a string")
    try:
        with open(args[0].value, "w", encoding="utf8") as file:
            file.write(args[1].value)
    except FileNotFoundError as e:
        raise RuntimeError(f"File '{args[0].value}' not found") from e
    return NullValue()


def _appendFile(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing filename or content argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Filename must be a string")
    if not isinstance(args[1], StringValue):
        raise RuntimeError("Content must be a string")
    try:
        with open(args[0].value, "a", encoding="utf8") as file:
            file.write(args[1].value)
    except FileNotFoundError as e:
        raise RuntimeError(f"File '{args[0].value}' not found") from e
    return NullValue()


def _round(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if len(args) == 1:
        if isinstance(args[0], NumberValue):
            return NumberValue(round(args[0].value))
    elif len(args) == 2:
        if isinstance(args[0], NumberValue) and isinstance(args[1], NumberValue):
            return NumberValue(round(args[0].value, int(args[1].value)))
    raise RuntimeError(f"Cannot round {args[0]}")


def _len(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if isinstance(args[0], StringValue):
        return NumberValue(len(args[0].value))

    if isinstance(args[0], ArrayValue):
        return NumberValue(len(args[0].elements))

    if isinstance(args[0], ObjectValue):
        return NumberValue(len(args[0].properties))

    raise RuntimeError(f"Cannot get length of {args[0]}")


def _exit(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if args and isinstance(args[0], NumberValue):
        sys.exit(int(args[0].value))
    sys.exit(0)


def _str(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return StringValue("")
    return StringValue(get_value(args[0]))


def _int(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if isinstance(args[0], (StringValue, NumberValue)):
        return NumberValue(int(args[0].value))
    raise RuntimeError(f"Cannot convert {args[0]} to int")


def _bool(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return BooleanValue(False)
    if isinstance(args[0], (StringValue, NumberValue, BooleanValue)):
        return BooleanValue(bool(args[0].value))
    if isinstance(args[0], ArrayValue):
        return BooleanValue(bool(args[0].elements))
    if isinstance(args[0], ObjectValue):
        return BooleanValue(bool(args[0].properties))
    raise RuntimeError(f"Cannot convert {args[0]} to bool")


def _array(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 1:
        if isinstance(args[0], StringValue):
            return ArrayValue([StringValue(char) for char in args[0].value])
        if isinstance(args[0], ObjectValue):
            return ArrayValue(list(args[0].properties.values()))
    return ArrayValue(args)


def _type(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    return StringValue(type(args[0]).__name__)


def _sum(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if isinstance(args[0], ArrayValue):
        if all(isinstance(i, NumberValue) for i in args[0].elements):
            return NumberValue(sum(i.value for i in args[0].elements))  # type: ignore
    raise RuntimeError(f"Cannot sum {args[0]}")


def _min(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if isinstance(args[0], ArrayValue):
        if all(isinstance(i, NumberValue) for i in args[0].elements):
            return NumberValue(min(i.value for i in args[0].elements))  # type: ignore
    raise RuntimeError(f"Cannot get min for {args[0]}")


def _max(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if isinstance(args[0], ArrayValue):
        if all(isinstance(i, NumberValue) for i in args[0].elements):
            return NumberValue(max(i.value for i in args[0].elements))  # type: ignore
    raise RuntimeError(f"Cannot get max for {args[0]}")


# Declare builtin modules
BUILTINS["file"] = ObjectValue(
    {
        "read": NativeFunctionValue(_readFile),
        "write": NativeFunctionValue(_writeFile),
        "append": NativeFunctionValue(_appendFile),
    },
    immutable=True,
)

BUILTINS["http"] = ObjectValue(
    {
        "get": NativeFunctionValue(_getRequest),
        "post": NativeFunctionValue(_postRequest),
    },
    immutable=True,
)

BUILTINS["math"] = ObjectValue(
    {
        "sum": NativeFunctionValue(_sum),
        "min": NativeFunctionValue(_min),
        "max": NativeFunctionValue(_max),
        "round": NativeFunctionValue(_round),
        "pi": NumberValue(math.pi),
        "sqrt": NativeFunctionValue(_sqrt),
        "random": NativeFunctionValue(_random),
    },
    immutable=True,
)

BUILTINS["time"] = ObjectValue({"time": NativeFunctionValue(_time)}, immutable=True)
